@@ -1,6 +1,7 @@
 //! The main context structure which drives the drawing process.
 
 use float_cmp::approx_eq;
+use gio::prelude::*;
 use glib::translate::*;
 use pango::ffi::PangoMatrix;
 use pango::prelude::FontMapExt;
@@ -12,7 +13,6 @@ use std::rc::Rc;
 use std::{borrow::Cow, sync::OnceLock};
 
 use crate::accept_language::UserLanguage;
-use crate::aspect_ratio::AspectRatio;
 use crate::bbox::BoundingBox;
 use crate::color::color_to_rgba;
 use crate::coord_units::CoordUnits;
@@ -24,7 +24,8 @@ use crate::filters::{self, FilterSpec};
 use crate::float_eq_cairo::ApproxEqCairo;
 use crate::gradient::{GradientVariant, SpreadMethod, UserSpaceGradient};
 use crate::layout::{
-    self, Filter, Image, Layer, LayerKind, Shape, StackingContext, Stroke, Text, TextSpan,
+    self, Filter, Group, Image, Layer, LayerKind, LayoutViewport, Shape, StackingContext, Stroke,
+    Text, TextSpan,
 };
 use crate::length::*;
 use crate::marker;
@@ -33,7 +34,7 @@ use crate::paint_server::{PaintSource, UserSpacePaintSource};
 use crate::path_builder::*;
 use crate::pattern::UserSpacePattern;
 use crate::properties::{
-    ClipRule, ComputedValues, FillRule, ImageRendering, MaskType, MixBlendMode, Opacity, Overflow,
+    ClipRule, ComputedValues, FillRule, ImageRendering, MaskType, MixBlendMode, Opacity,
     PaintTarget, ShapeRendering, StrokeLinecap, StrokeLinejoin, TextRendering,
 };
 use crate::rect::{rect_to_transform, IRect, Rect};
@@ -54,12 +55,6 @@ pub struct FontOptions {
     options: cairo::FontOptions,
 }
 
-#[derive(Debug, Copy, Clone, PartialEq)]
-pub enum ClipMode {
-    ClipToViewport,
-    NoClip,
-}
-
 /// Set path on the cairo context, or clear it.
 /// This helper object keeps track whether the path has been set already,
 /// so that it isn't recalculated every so often.
@@ -166,12 +161,21 @@ impl Viewport {
             transform: self.transform,
         }
     }
+
+    pub fn with_composed_transform(&self, transform: Transform) -> Viewport {
+        Viewport {
+            dpi: self.dpi,
+            vbox: self.vbox,
+            transform: self.transform.pre_transform(&transform),
+        }
+    }
 }
 
 /// Values that stay constant during rendering with a DrawingCtx.
 #[derive(Clone)]
 pub struct RenderingConfiguration {
     pub dpi: Dpi,
+    pub cancellable: Option<gio::Cancellable>,
     pub user_language: UserLanguage,
     pub svg_nesting: SvgNesting,
     pub measuring: bool,
@@ -277,7 +281,11 @@ pub fn draw_tree(
 
     user_bbox.insert(&content_bbox);
 
-    Ok(user_bbox)
+    if draw_ctx.is_rendering_cancelled() {
+        Err(InternalRenderingError::Cancelled)
+    } else {
+        Ok(user_bbox)
+    }
 }
 
 pub fn with_saved_cr<O, F>(cr: &cairo::Context, f: F) -> Result<O, InternalRenderingError>
@@ -350,6 +358,7 @@ impl DrawingCtx {
     pub fn rendering_options(&self, svg_nesting: SvgNesting) -> RenderingOptions {
         RenderingOptions {
             dpi: self.config.dpi,
+            cancellable: self.config.cancellable.clone(),
             user_language: self.config.user_language.clone(),
             svg_nesting,
             testing: self.config.testing,
@@ -455,17 +464,21 @@ impl DrawingCtx {
     pub fn push_new_viewport(
         &self,
         current_viewport: &Viewport,
-        vbox: Option<ViewBox>,
-        viewport_rect: Rect,
-        preserve_aspect_ratio: AspectRatio,
-        clip_mode: ClipMode,
+        layout_viewport: &LayoutViewport,
     ) -> Option<Viewport> {
-        if let ClipMode::ClipToViewport = clip_mode {
-            clip_to_rectangle(&self.cr, &viewport_rect);
+        let LayoutViewport {
+            geometry,
+            vbox,
+            preserve_aspect_ratio,
+            overflow,
+        } = *layout_viewport;
+
+        if !overflow.overflow_allowed() || (vbox.is_some() && preserve_aspect_ratio.is_slice()) {
+            clip_to_rectangle(&self.cr, &geometry);
         }
 
         preserve_aspect_ratio
-            .viewport_to_viewbox_transform(vbox, &viewport_rect)
+            .viewport_to_viewbox_transform(vbox, &geometry)
             .unwrap_or_else(|_e| {
                 match vbox {
                     None => unreachable!(
@@ -485,12 +498,13 @@ impl DrawingCtx {
                 None
             })
             .map(|t| {
+                // FMQ: here
                 self.cr.transform(t.into());
 
                 Viewport {
                     dpi: self.config.dpi,
                     vbox: vbox.unwrap_or(current_viewport.vbox),
-                    transform: current_viewport.transform.post_transform(&t),
+                    transform: current_viewport.transform.pre_transform(&t),
                 }
             })
     }
@@ -517,6 +531,7 @@ impl DrawingCtx {
             let transform_for_clip = ValidTransform::try_from(node_transform)?;
 
             let orig_transform = self.get_transform();
+            // FMQ: here
             self.cr.transform(transform_for_clip.into());
 
             for child in node.children().filter(|c| {
@@ -614,6 +629,7 @@ impl DrawingCtx {
                 mask_cr.transform(ValidTransform::try_from(bbtransform)?.into());
             }
 
+            // FMQ: above - and here, the mask_viewport need the new bbtransform composed too
             let mask_viewport = viewport.with_units(mask.get_content_units());
 
             let mut mask_draw_ctx = self.nested(mask_cr);
@@ -633,8 +649,11 @@ impl DrawingCtx {
                 &stacking_ctx,
                 acquired_nodes,
                 &mask_viewport,
+                None,
                 false,
-                &mut |an, dc| mask_node.draw_children(an, &cascaded, &mask_viewport, dc, false),
+                &mut |an, dc, new_viewport| {
+                    mask_node.draw_children(an, &cascaded, new_viewport, dc, false)
+                },
             );
 
             rsvg_log!(self.session, ")");
@@ -654,24 +673,56 @@ impl DrawingCtx {
         Ok(Some(mask))
     }
 
+    fn is_rendering_cancelled(&self) -> bool {
+        match &self.config.cancellable {
+            None => false,
+            Some(cancellable) => cancellable.is_cancelled(),
+        }
+    }
+
     pub fn with_discrete_layer(
         &mut self,
         stacking_ctx: &StackingContext,
         acquired_nodes: &mut AcquiredNodes<'_>,
         viewport: &Viewport,
+        layout_viewport: Option<LayoutViewport>,
         clipping: bool,
         draw_fn: &mut dyn FnMut(
             &mut AcquiredNodes<'_>,
             &mut DrawingCtx,
+            &Viewport,
         ) -> Result<BoundingBox, InternalRenderingError>,
     ) -> Result<BoundingBox, InternalRenderingError> {
+        if self.is_rendering_cancelled() {
+            return Err(InternalRenderingError::Cancelled);
+        }
+
         let stacking_ctx_transform = ValidTransform::try_from(stacking_ctx.transform)?;
 
         let orig_transform = self.get_transform();
+
+        // See the comment below about "not using that transform anywhere" (the viewport's).
+        let viewport = viewport.with_composed_transform(stacking_ctx.transform);
         self.cr.transform(stacking_ctx_transform.into());
 
         let res = if clipping {
-            draw_fn(acquired_nodes, self)
+            if let Some(layout_viewport) = layout_viewport.as_ref() {
+                // FIXME: here we ignore the Some() result of push_new_viewport().  We do that because
+                // the returned one is just a copy of the one that got passeed in, but with a changed
+                // transform.  However, we are in fact not using that transform anywhere!
+                //
+                // In case push_new_viewport() returns None, we just don't draw anything.
+                //
+                // Note that push_new_viewport() changes the cr's transform.  However it will be restored
+                // at the end of this function with set_matrix.
+                if let Some(new_viewport) = self.push_new_viewport(&viewport, layout_viewport) {
+                    draw_fn(acquired_nodes, self, &new_viewport)
+                } else {
+                    Ok(self.empty_bbox())
+                }
+            } else {
+                draw_fn(acquired_nodes, self, &viewport)
+            }
         } else {
             with_saved_cr(&self.cr.clone(), || {
                 if let Some(ref link_target) = stacking_ctx.link_target {
@@ -690,7 +741,7 @@ impl DrawingCtx {
                 self.clip_to_node(
                     &stacking_ctx.clip_in_user_space,
                     acquired_nodes,
-                    viewport,
+                    &viewport,
                     &self.empty_bbox(),
                 )?;
 
@@ -720,11 +771,28 @@ impl DrawingCtx {
                     cr.set_matrix(ValidTransform::try_from(affines.for_temporary_surface)?.into());
 
                     let (source_surface, mut res, bbox) = {
-                        let mut temporary_draw_ctx = self.nested(cr);
+                        let mut temporary_draw_ctx = self.nested(cr.clone());
 
                         // Draw!
 
-                        let res = draw_fn(acquired_nodes, &mut temporary_draw_ctx);
+                        let res = with_saved_cr(&cr, || {
+                            if let Some(layout_viewport) = layout_viewport.as_ref() {
+                                // FIXME: here we ignore the Some() result of push_new_viewport().  We do that because
+                                // the returned one is just a copy of the one that got passeed in, but with a changed
+                                // transform.  However, we are in fact not using that transform anywhere!
+                                //
+                                // In case push_new_viewport() returns None, we just don't draw anything.
+                                if let Some(new_viewport) =
+                                    temporary_draw_ctx.push_new_viewport(&viewport, layout_viewport)
+                                {
+                                    draw_fn(acquired_nodes, &mut temporary_draw_ctx, &new_viewport)
+                                } else {
+                                    Ok(self.empty_bbox())
+                                }
+                            } else {
+                                draw_fn(acquired_nodes, &mut temporary_draw_ctx, &viewport)
+                            }
+                        });
 
                         let bbox = if let Ok(ref bbox) = res {
                             *bbox
@@ -741,13 +809,13 @@ impl DrawingCtx {
                             let stroke_paint_source =
                                 Rc::new(filter.stroke_paint_source.to_user_space(
                                     &bbox.rect,
-                                    viewport,
+                                    &viewport,
                                     &filter.normalize_values,
                                 ));
                             let fill_paint_source =
                                 Rc::new(filter.fill_paint_source.to_user_space(
                                     &bbox.rect,
-                                    viewport,
+                                    &viewport,
                                     &filter.normalize_values,
                                 ));
 
@@ -762,7 +830,7 @@ impl DrawingCtx {
 
                             let filtered_surface = temporary_draw_ctx
                                 .run_filters(
-                                    viewport,
+                                    &viewport,
                                     surface_to_filter,
                                     filter,
                                     acquired_nodes,
@@ -796,7 +864,7 @@ impl DrawingCtx {
                     self.clip_to_node(
                         &stacking_ctx.clip_in_object_space,
                         acquired_nodes,
-                        viewport,
+                        &viewport,
                         &bbox,
                     )?;
 
@@ -806,7 +874,7 @@ impl DrawingCtx {
                         res = res.and_then(|bbox| {
                             self.generate_cairo_mask(
                                 mask_node,
-                                viewport,
+                                &viewport,
                                 affines.for_temporary_surface,
                                 &bbox,
                                 acquired_nodes,
@@ -845,8 +913,23 @@ impl DrawingCtx {
 
                     self.cr.set_matrix(affine_at_start.into());
                     res
+                } else if let Some(layout_viewport) = layout_viewport.as_ref() {
+                    // FIXME: here we ignore the Some() result of push_new_viewport().  We do that because
+                    // the returned one is just a copy of the one that got passeed in, but with a changed
+                    // transform.  However, we are in fact not using that transform anywhere!
+                    //
+                    // In case push_new_viewport() returns None, we just don't draw anything.
+                    //
+                    // Note that push_new_viewport() changes the cr's transform.  However it will be restored
+                    // at the end of this function with set_matrix.
+                    if let Some(new_viewport) = self.push_new_viewport(&viewport, layout_viewport) {
+                        draw_fn(acquired_nodes, self, &new_viewport)
+                    } else {
+                        self.cr.set_matrix(orig_transform.into());
+                        Ok(self.empty_bbox())
+                    }
                 } else {
-                    draw_fn(acquired_nodes, self)
+                    draw_fn(acquired_nodes, self, &viewport)
                 };
 
                 if stacking_ctx.link_target.is_some() {
@@ -1093,12 +1176,13 @@ impl DrawingCtx {
                         &stacking_ctx,
                         acquired_nodes,
                         &pattern_viewport,
+                        None,
                         false,
-                        &mut |an, dc| {
+                        &mut |an, dc, new_viewport| {
                             pattern_node.draw_children(
                                 an,
                                 &pattern_cascaded,
-                                &pattern_viewport,
+                                new_viewport,
                                 dc,
                                 false,
                             )
@@ -1233,6 +1317,13 @@ impl DrawingCtx {
                 clipping,
                 viewport,
             ),
+            LayerKind::Group(group) => self.draw_group(
+                group,
+                &layer.stacking_ctx,
+                acquired_nodes,
+                clipping,
+                viewport,
+            ),
         }
     }
 
@@ -1263,8 +1354,9 @@ impl DrawingCtx {
             stacking_ctx,
             acquired_nodes,
             viewport,
+            None,
             clipping,
-            &mut |an, dc| {
+            &mut |an, dc, new_viewport| {
                 let cr = dc.cr.clone();
 
                 let transform = dc.get_transform_for_stacking_ctx(stacking_ctx, clipping)?;
@@ -1323,7 +1415,11 @@ impl DrawingCtx {
                             PaintTarget::Markers => {
                                 path_helper.unset();
                                 marker::render_markers_for_shape(
-                                    shape, viewport, dc, an, clipping,
+                                    shape,
+                                    new_viewport,
+                                    dc,
+                                    an,
+                                    clipping,
                                 )?;
                             }
                         }
@@ -1384,44 +1480,33 @@ impl DrawingCtx {
         let image_height = f64::from(image_height);
         let vbox = ViewBox::from(Rect::from_size(image_width, image_height));
 
-        let clip_mode = if !(image.overflow == Overflow::Auto
-            || image.overflow == Overflow::Visible)
-            && image.aspect.is_slice()
-        {
-            ClipMode::ClipToViewport
-        } else {
-            ClipMode::NoClip
-        };
-
         // The bounding box for <image> is decided by the values of the image's x, y, w, h
         // and not by the final computed image bounds.
         let bounds = self.empty_bbox().with_rect(image.rect);
 
+        let layout_viewport = LayoutViewport {
+            vbox: Some(vbox),
+            geometry: image.rect,
+            preserve_aspect_ratio: image.aspect,
+            overflow: image.overflow,
+        };
+
         if image.is_visible {
             self.with_discrete_layer(
                 stacking_ctx,
                 acquired_nodes,
-                viewport, // FIXME: should this be the push_new_viewport below?
+                viewport,
+                Some(layout_viewport),
                 clipping,
-                &mut |_an, dc| {
-                    with_saved_cr(&dc.cr.clone(), || {
-                        if let Some(_params) = dc.push_new_viewport(
-                            viewport,
-                            Some(vbox),
-                            image.rect,
-                            image.aspect,
-                            clip_mode,
-                        ) {
-                            dc.paint_surface(
-                                &image.surface,
-                                image_width,
-                                image_height,
-                                image.image_rendering,
-                            )?;
-                        }
+                &mut |_an, dc, _new_viewport| {
+                    dc.paint_surface(
+                        &image.surface,
+                        image_width,
+                        image_height,
+                        image.image_rendering,
+                    )?;
 
-                        Ok(bounds)
-                    })
+                    Ok(bounds)
                 },
             )
         } else {
@@ -1429,6 +1514,17 @@ impl DrawingCtx {
         }
     }
 
+    fn draw_group(
+        &mut self,
+        _group: &Group,
+        _stacking_ctx: &StackingContext,
+        _acquired_nodes: &mut AcquiredNodes<'_>,
+        _clipping: bool,
+        _viewport: &Viewport,
+    ) -> Result<BoundingBox, InternalRenderingError> {
+        unimplemented!()
+    }
+
     fn draw_text_span(
         &mut self,
         span: &TextSpan,
@@ -1538,8 +1634,9 @@ impl DrawingCtx {
             stacking_ctx,
             acquired_nodes,
             viewport,
+            None,
             clipping,
-            &mut |an, dc| {
+            &mut |an, dc, _new_viewport| {
                 let mut bbox = dc.empty_bbox();
 
                 for span in &text.spans {
@@ -1587,7 +1684,7 @@ impl DrawingCtx {
                 );
 
                 cr.set_matrix(ValidTransform::try_from(affines.for_snapshot)?.into());
-                cr.set_source_surface(&draw.target(), 0.0, 0.0)?;
+                cr.set_source_surface(draw.target(), 0.0, 0.0)?;
                 cr.paint()?;
             }
 
@@ -1734,6 +1831,7 @@ impl DrawingCtx {
 
         let orig_transform = self.get_transform();
 
+        // FMQ: here
         self.cr
             .transform(ValidTransform::try_from(values.transform())?.into());
 
@@ -1749,7 +1847,7 @@ impl DrawingCtx {
             None
         };
 
-        let res = if let Some((viewbox, preserve_aspect_ratio)) = defines_a_viewport {
+        let res = if let Some((vbox, preserve_aspect_ratio)) = defines_a_viewport {
             // <symbol> and <svg> define a viewport, as described in the specification:
             // https://www.w3.org/TR/SVG2/struct.html#UseElement
             // https://gitlab.gnome.org/GNOME/librsvg/-/issues/875#note_1482705
@@ -1758,13 +1856,6 @@ impl DrawingCtx {
 
             let child_values = elt.get_computed_values();
 
-            // FIXME: do we need to look at preserveAspectRatio.slice, like in draw_image()?
-            let clip_mode = if !child_values.is_overflow() {
-                ClipMode::ClipToViewport
-            } else {
-                ClipMode::NoClip
-            };
-
             let stacking_ctx = StackingContext::new(
                 self.session(),
                 acquired_nodes,
@@ -1774,34 +1865,32 @@ impl DrawingCtx {
                 values,
             );
 
+            let layout_viewport = LayoutViewport {
+                vbox,
+                geometry: use_rect,
+                preserve_aspect_ratio,
+                overflow: child_values.overflow(),
+            };
+
             self.with_discrete_layer(
                 &stacking_ctx,
                 acquired_nodes,
-                viewport, // FIXME: should this be the child_viewport from below?
+                viewport,
+                Some(layout_viewport),
                 clipping,
-                &mut |an, dc| {
-                    if let Some(child_viewport) = dc.push_new_viewport(
-                        viewport,
-                        viewbox,
-                        use_rect,
-                        preserve_aspect_ratio,
-                        clip_mode,
-                    ) {
-                        child.draw_children(
-                            an,
-                            &CascadedValues::new_from_values(
-                                child,
-                                values,
-                                Some(fill_paint.clone()),
-                                Some(stroke_paint.clone()),
-                            ),
-                            &child_viewport,
-                            dc,
-                            clipping,
-                        )
-                    } else {
-                        Ok(dc.empty_bbox())
-                    }
+                &mut |an, dc, new_viewport| {
+                    child.draw_children(
+                        an,
+                        &CascadedValues::new_from_values(
+                            child,
+                            values,
+                            Some(fill_paint.clone()),
+                            Some(stroke_paint.clone()),
+                        ),
+                        new_viewport,
+                        dc,
+                        clipping,
+                    )
                 },
             )
         } else {
@@ -1820,8 +1909,9 @@ impl DrawingCtx {
                 &stacking_ctx,
                 acquired_nodes,
                 viewport,
+                None,
                 clipping,
-                &mut |an, dc| {
+                &mut |an, dc, new_viewport| {
                     child.draw(
                         an,
                         &CascadedValues::new_from_values(
@@ -1830,7 +1920,7 @@ impl DrawingCtx {
                             Some(fill_paint.clone()),
                             Some(stroke_paint.clone()),
                         ),
-                        viewport,
+                        new_viewport,
                         dc,
                         clipping,
                     )
@@ -1865,20 +1955,6 @@ impl DrawingCtx {
     }
 }
 
-pub fn compute_path_extents(path: &Path) -> Result<Option<Rect>, InternalRenderingError> {
-    if path.is_empty() {
-        return Ok(None);
-    }
-
-    let surface = cairo::RecordingSurface::create(cairo::Content::ColorAlpha, None)?;
-    let cr = cairo::Context::new(&surface)?;
-
-    path.to_cairo(&cr, false)?;
-    let (x0, y0, x1, y1) = cr.path_extents()?;
-
-    Ok(Some(Rect::new(x0, y0, x1, y1)))
-}
-
 impl From<ImageRendering> for Interpolation {
     fn from(r: ImageRendering) -> Interpolation {
         match r {
@@ -1894,6 +1970,20 @@ impl From<ImageRendering> for Interpolation {
     }
 }
 
+pub fn compute_path_extents(path: &Path) -> Result<Option<Rect>, InternalRenderingError> {
+    if path.is_empty() {
+        return Ok(None);
+    }
+
+    let surface = cairo::RecordingSurface::create(cairo::Content::ColorAlpha, None)?;
+    let cr = cairo::Context::new(&surface)?;
+
+    path.to_cairo(&cr, false)?;
+    let (x0, y0, x1, y1) = cr.path_extents()?;
+
+    Ok(Some(Rect::new(x0, y0, x1, y1)))
+}
+
 /// Create a Pango context with a particular configuration.
 pub fn create_pango_context(font_options: &FontOptions, transform: &Transform) -> pango::Context {
     let font_map = pangocairo::FontMap::default();
