@@ -95,13 +95,6 @@ pub enum DefsLookupErrorKind {
     /// Error when parsing the id to lookup.
     InvalidId,
 
-    /// Used when the public API tries to look up an external URL, which is not allowed.
-    ///
-    /// This catches the case where a public API wants to be misused to access an external
-    /// resource.  For example, `SvgHandle.has_sub("https://evil.com/phone_home#element_id")` will
-    /// fail with this error.
-    CannotLookupExternalReferences,
-
     /// For internal use only.
     ///
     // FIXME: this is returned internally from Handle.lookup_node(), and gets translated
@@ -113,9 +106,6 @@ impl fmt::Display for DefsLookupErrorKind {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match *self {
             DefsLookupErrorKind::InvalidId => write!(f, "invalid id"),
-            DefsLookupErrorKind::CannotLookupExternalReferences => {
-                write!(f, "cannot lookup references to elements in external files")
-            }
             DefsLookupErrorKind::NotFound => write!(f, "not found"),
         }
     }
@@ -131,7 +121,7 @@ impl fmt::Display for DefsLookupErrorKind {
 ///   "If a transform function causes the current transformation matrix of an
 ///   object to be non-invertible, the object and its content do not get
 ///   displayed."
-#[derive(Debug, Clone)]
+#[derive(Clone)]
 pub enum InternalRenderingError {
     /// An error from the rendering backend.
     Rendering(String),
@@ -145,6 +135,8 @@ pub enum InternalRenderingError {
     /// the problematic element.
     InvalidTransform,
 
+    CircularReference(Node),
+
     /// Tried to reference an SVG element that does not exist.
     IdNotFound,
 
@@ -153,6 +145,9 @@ pub enum InternalRenderingError {
 
     /// Not enough memory was available for rendering.
     OutOfMemory(String),
+
+    /// The rendering was interrupted via a [`gio::Cancellable`].
+    Cancelled,
 }
 
 impl From<DefsLookupErrorKind> for InternalRenderingError {
@@ -176,9 +171,13 @@ impl fmt::Display for InternalRenderingError {
             InternalRenderingError::Rendering(ref s) => write!(f, "rendering error: {s}"),
             InternalRenderingError::LimitExceeded(ref l) => write!(f, "{l}"),
             InternalRenderingError::InvalidTransform => write!(f, "invalid transform"),
+            InternalRenderingError::CircularReference(ref c) => {
+                write!(f, "circular reference in element {c}")
+            }
             InternalRenderingError::IdNotFound => write!(f, "element id not found"),
             InternalRenderingError::InvalidId(ref s) => write!(f, "invalid id: {s:?}"),
             InternalRenderingError::OutOfMemory(ref s) => write!(f, "out of memory: {s}"),
+            InternalRenderingError::Cancelled => write!(f, "rendering cancelled"),
         }
     }
 }
