@@ -13,9 +13,10 @@ use crate::element::{set_attribute, ElementTrait};
 use crate::error::*;
 use crate::iri::Iri;
 use crate::is_element_of_type;
-use crate::layout::{Layer, LayerKind, Marker, Shape, StackingContext, Stroke};
+use crate::layout::{self, Layer, LayerKind, Marker, Shape, StackingContext, Stroke};
 use crate::length::*;
 use crate::node::{CascadedValues, Node, NodeBorrow};
+use crate::paint_server::PaintSource;
 use crate::parsers::{optional_comma, Parse, ParseValue};
 use crate::path_builder::{LargeArc, Path as SvgPath, PathBuilder, Sweep};
 use crate::properties::ComputedValues;
@@ -44,6 +45,31 @@ trait BasicShape {
     fn make_shape(&self, params: &NormalizeParams, values: &ComputedValues) -> ShapeDef;
 }
 
+fn validate_path(
+    path: &Rc<SvgPath>,
+    viewport: &Viewport,
+    normalize_values: &NormalizeValues,
+    stroke_paint: &PaintSource,
+    fill_paint: &PaintSource,
+) -> Result<layout::Path, InternalRenderingError> {
+    if path.has_unsuitable_coordinates(&viewport.transform) {
+        return Ok(layout::Path::Invalid(String::from(
+            "path has coordinates that are unsuitable for Cairo",
+        )));
+    }
+
+    let extents = compute_path_extents(path)?;
+    let stroke_paint = stroke_paint.to_user_space(&extents, viewport, normalize_values);
+    let fill_paint = fill_paint.to_user_space(&extents, viewport, normalize_values);
+
+    Ok(layout::Path::Validated {
+        path: Rc::clone(path),
+        extents,
+        stroke_paint,
+        fill_paint,
+    })
+}
+
 fn draw_basic_shape(
     basic_shape: &dyn BasicShape,
     node: &Node,
@@ -115,21 +141,25 @@ fn draw_basic_shape(
         context_fill: fill_paint.clone(),
     };
 
-    let extents = compute_path_extents(&shape_def.path)?;
-
     let normalize_values = NormalizeValues::new(values);
 
-    let stroke_paint = stroke_paint.to_user_space(&extents, viewport, &normalize_values);
-    let fill_paint = fill_paint.to_user_space(&extents, viewport, &normalize_values);
+    let path = validate_path(
+        &shape_def.path,
+        viewport,
+        &normalize_values,
+        &stroke_paint,
+        &fill_paint,
+    )?;
+
+    if let layout::Path::Invalid(ref reason) = path {
+        rsvg_log!(session, "will not render {node}: {reason}");
+    }
 
     let shape = Box::new(Shape {
-        path: shape_def.path,
-        extents,
+        path,
         is_visible,
         paint_order,
         stroke,
-        stroke_paint,
-        fill_paint,
         fill_rule,
         clip_rule,
         shape_rendering,
